"""
Implements the game and the necessary input/output logic for interacting with the user.
"""
from .game_model import Minefield, random_minefield, GameState, CellState
from .keyboard_listener import input_loop
from .renderer import render



"""
An input loop based around click.getchar(). Supports arrow key escape sequences on both Unix-like platforms and Windows.

When run directly this this script will enter a test loop.
"""

from enum import Enum

import click


class ArrowKeyMapping(Enum):
    UP = "w"
    DOWN = "s"
    LEFT = "a"
    RIGHT = "d"


def input_loop(handler_func):
    """
    Loops until ESC is pressed or the program is killed. Calls handler_func each time a keystroke is received.
    """
    while True:
        try:
            ch = click.getchar()
        except (KeyboardInterrupt, EOFError):
            return

        # If it's an escape sequence grab the ascii char code
        if len(ch) > 1:
            escape_ord = ord(ch[0])
            ch = ch[-1]
        else:
            escape_ord = None

        if 32 <= ord(ch) <= 126:
            # Only pass "regular" keys onto the handler

            if escape_ord == 27 or escape_ord == 224:
                # Translate the arrow key escape sequence into the mapped letter
                if ch == "A" or ch == "H":
                    ch = ArrowKeyMapping.UP.value
                elif ch == "B" or ch == "P":
                    ch = ArrowKeyMapping.DOWN.value
                elif ch == "D" or ch == "K":
                    ch = ArrowKeyMapping.LEFT.value
                elif ch == "C" or ch == "M":
                    ch = ArrowKeyMapping.RIGHT.value

            handler_func(ch.lower())
        elif ord(ch) == 13:
            # Translate the enter key to a newline
            handler_func("\n")
        elif ord(ch) == 27:
            # Exit on ESC
            return


def demo_handler(key):
    print("Processing {}".format(repr(key)))


if __name__ == "__main__":
    input_loop(demo_handler)



"""
Entry point and CLI implementation for terminal-mines.
"""

import click

from .game_logic import random_minefield, Minefield, GameState, input_loop, render

DIFFICULTY_PRESETS = {
    "easy": (10, 8, 8),
    "intermediate": (40, 16, 16),
    "expert": (99, 16, 30)
}


class DifficultyParamType(click.ParamType):
    """
    Converts the provided difficulty string into the 3 args expected by random_minefield().
    """
    def convert(self, value, param, ctx):
        if value in DIFFICULTY_PRESETS:
            return DIFFICULTY_PRESETS[value]
        elif "," not in value:
            self.fail("'{}' is not a valid difficulty name".format(value), param, ctx)
        else:
            try:
                args = tuple(map(int, value.split(",")))

                if len(args) != 3:
                    raise ValueError
                elif args[0] < 0 or args[1] < 0 or args[2] < 0:
                    raise ValueError
                elif args[1] > 50 or args[2] > 50:
                    self.fail("the game board cannot be larger than 50 cells on either side", param, ctx)
                elif args[0] > args[1] * args[2]:
                    self.fail("{} mines cannot fit in a board size of {} by {}".format(*args), param, ctx)

                return args
            except ValueError:
                self.fail("a custom difficulty must be made of 3 positive integers separated by commas", param, ctx)


@click.command()
@click.pass_context
@click.argument("difficulty", default="easy", type=DifficultyParamType())
@click.option("mines_file", "--mines", type=click.File(), help="File containing custom mine placements.")
def main(ctx, difficulty, mines_file):
    """
    Terminal Mines

    A command-line clone of Minesweeper in Python.

    \b
    Controls:
    - WASD or arrow keys to move the cursor
    - Enter or space to reveal the current cell
    - e or ' to place a flag
    - ESC to quit

    DIFFICULTY can either be "easy", "intermediate", "expert" or a custom difficulty of the form
    "<number of mines>,<width>,<height>". If no difficulty is specified Terminal Mines will default to easy.

    The mines file (if provided) is used to control the placement of mines. It must be a CSV where each line is of the
    form "<x>,<y>". Both coordinates are 0-based and count from the top-left corner of the game board. If any of the
    specified mines are outside the bounds of the game board they will be skipped. If a mines file is provided the
    "number of mines" portion of the difficulty setting will be ignored.
    """
    if mines_file:
        mines = set(map(lambda line: line.strip(), mines_file))
        minefield = Minefield(difficulty[1], difficulty[2], mines)

        if minefield.num_mines == 0:
            ctx.fail("Mines file did not contain any valid mines")
    else:
        minefield = random_minefield(*difficulty)

    def handle_key(key):
        if key == "w":
            minefield.y = (minefield.y - 1) % minefield.height
        elif key == "s":
            minefield.y = (minefield.y + 1) % minefield.height
        elif key == "a":
            minefield.x = (minefield.x - 1) % minefield.width
        elif key == "d":
            minefield.x = (minefield.x + 1) % minefield.width
        elif key == "e" or key == "'":
            minefield.flag_cell(minefield.x, minefield.y)
        elif key == "\n" or key == " ":
            minefield.reveal_cell(minefield.x, minefield.y)

        render(minefield)

        if minefield.state != GameState.IN_PROGRESS:
            ctx.exit(0)

    render(minefield)
    input_loop(handle_key)



"""
Handles the rendering of the game state to the console.
"""

from itertools import chain

from click import clear, style, echo, get_current_context

from .game_model import GameState, CellState

fg_mapping = {
    CellState.FLAGGED: "bright_green",
    CellState.WARN1: "bright_cyan",
    CellState.WARN2: "cyan",
    CellState.WARN3: "bright_blue",
    CellState.WARN4: "bright_magenta",
    CellState.WARN5: "magenta",
    CellState.WARN6: "bright_yellow",
    CellState.WARN7: "red",
    CellState.WARN8: "red",
    CellState.EXPLODED: "bright_red"
}


def render(minefield):
    """
    Clears the screen and renders the current game state.
    """
    clear()

    def render_cell(iter_x, iter_y):
        cell = minefield.get_cell(iter_x, iter_y)

        fg = fg_mapping.get(cell.state, None)

        if iter_x == minefield.x and iter_y == minefield.y:
            bg = "green"
            fg = "black"    # Override the foreground color to make it more readable against the green background
        elif minefield.state != GameState.IN_PROGRESS and cell.state == CellState.FLAGGED:
            bg = None if cell.is_mine else "red"    # Indicates incorrectly placed flag
        else:
            bg = None

        return style(cell.state.value, bg=bg, fg=fg)

    def gen_lines():
        yield chr(0x250C) + chr(0x2500) * (minefield.width * 2 + 1) + chr(0x2510)

        for iter_y in range(minefield.height):
            iter_cells = (render_cell(iter_x, iter_y) for iter_x in range(minefield.width))
            yield " ".join(chain(chr(0x2502), iter_cells, chr(0x2502)))

        yield chr(0x2514) + chr(0x2500) * (minefield.width * 2 + 1) + chr(0x2518)

        if minefield.state == GameState.WON:
            yield " Game won"
        elif minefield.state == GameState.LOST:
            yield " Game lost"
        else:
            yield " Flags remaining: {}".format(minefield.flags_remaining)

    try:
        echo("\n".join(gen_lines()))
    except UnicodeEncodeError:
        # The Git bash emulator on Windows doesn't play nice with unicode or the input loop. To save the user from this
        #   we'll quit while we're ahead.
        get_current_context().fail("terminal-mines does not support the Git bash emulator on Windows. Please use CMD "
                                   "or PowerShell instead.")



from setuptools import setup, find_packages

with open("README.md") as file:
    # Long description is the readme minus the header line
    long_description = file.read()[17:]

setup(
    name="terminal-mines",
    version="1.1",
    python_requires="~=3.4",
    license="MIT",
    author="Joel Eager",
    description="A command-line clone of Minesweeper in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/JoelEager/terminal-mines",
    packages=find_packages(),
    install_requires=[
        "click==7.0",
    ],
    entry_points="""
        [console_scripts]
        mines=terminal_mines.mines:main
    """,
    classifiers=[
        "Topic :: Games/Entertainment :: Puzzle Games",
        "Environment :: Console",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows"
    ]
)



"""
Models the game state and exposes functions for manipulating it.
"""

from enum import Enum
from random import randint


class CellState(Enum):
    UNKNOWN = "?"
    SAFE = "-"
    WARN1 = "1"
    WARN2 = "2"
    WARN3 = "3"
    WARN4 = "4"
    WARN5 = "5"
    WARN6 = "6"
    WARN7 = "7"
    WARN8 = "8"
    FLAGGED = "F"
    EXPLODED = "X"


class GameState(Enum):
    IN_PROGRESS = 0
    WON = 1
    LOST = 2


class Cell:
    """
    Represents one cell on the game board. The state is shown to the user while the is_mine value is used internally.
    """
    def __init__(self, is_mine):
        self.is_mine = is_mine
        self.state = CellState.UNKNOWN

    def __repr__(self):
        return "{}({}, {})".format(type(self).__name__, self.is_mine, self.state.value)


class Minefield:
    """
    Stores the state of the game and the position of the cursor. Provides functions for interacting with the game.
    """
    def __init__(self, width, height, mines):
        """
        The mines arg must be a set of strings of the form "x,y".
        """
        self.width = width
        self.height = height

        self.x = 0      # The x cord of the currently selected cell
        self.y = 0      # The y cord of the currently selected cell
        self.state = GameState.IN_PROGRESS

        self.rows = [[Cell("{},{}".format(x, y) in mines) for x in range(width)] for y in range(height)]

    def __repr__(self):
        return "{}({}, {})".format(type(self).__name__, self.width, self.height)

    @property
    def cells(self):
        for row in self.rows:
            for cell in row:
                yield cell

    @property
    def num_mines(self):
        return len([cell for cell in self.cells if cell.is_mine])

    @property
    def flags_remaining(self):
        return self.num_mines - len([cell for cell in self.cells if cell.state == CellState.FLAGGED])

    def get_cell(self, x, y):
        if 0 <= x < self.width and 0 <= y < self.height:
            return self.rows[y][x]
        else:
            raise IndexError

    def neighbors(self, x, y):
        """
        Iterates over neighboring cells
        """
        for offset_x, offset_y in ((-1, -1), (-1, 0), (-1, 1), (0, -1), (0, 1), (1, -1), (1, 0), (1, 1)):
            try:
                yield self.get_cell(x + offset_x, y + offset_y)
            except IndexError:
                continue

    def reveal_cell(self, x, y):
        """
        Reveals the given cell and updates the game state. Will recursively reveal other cells if the given one is safe.
        """
        target = self.get_cell(x, y)

        if target.state != CellState.UNKNOWN:
            return

        if target.is_mine:
            # Game lost; update all un-flagged mines as exploded
            target.state = CellState.EXPLODED
            
            for cell in self.cells:
                if cell.state == CellState.UNKNOWN and cell.is_mine:
                    cell.state = CellState.EXPLODED
                        
            self.state = GameState.LOST
        else:
            neighbor_mines = len([cell for cell in self.neighbors(x, y) if cell.is_mine])

            if neighbor_mines == 0:
                target.state = CellState.SAFE

                # Use recursion to propagate the reveal to neighboring cells
                for offset_x, offset_y in ((-1, -1), (-1, 0), (-1, 1), (0, -1), (0, 1), (1, -1), (1, 0), (1, 1)):
                    try:
                        self.reveal_cell(x + offset_x, y + offset_y)
                    except IndexError:
                        continue
            else:
                target.state = CellState(str(neighbor_mines))
                
    def flag_cell(self, x, y):
        """
        Toggles a cell between the unknown and flagged states. Does nothing if called on a revealed cell or if the
        player is out of flags.
        """
        target = self.get_cell(x, y)

        if target.state == CellState.FLAGGED:
            target.state = CellState.UNKNOWN
        elif target.state == CellState.UNKNOWN and self.flags_remaining > 0:
            target.state = CellState.FLAGGED

            # Check if the game has been won
            for cell in self.cells:
                if cell.is_mine and cell.state != CellState.FLAGGED:
                    return
                elif not cell.is_mine and cell.state == CellState.FLAGGED:
                    return

            self.state = GameState.WON


def random_minefield(num_mines, width, height):
    """
    :return: A new Minefield instance with a random set of mines.
    """
    mines = set()

    while len(mines) != num_mines:
        mines.add("{},{}".format(randint(0, width - 1), randint(0, height - 1)))

    return Minefield(width, height, mines)
